@@ -4,14 +4,16 @@ use 5.008006;
 use strict;
 use warnings;
 
-use JSON;
+use JSON::XS;
 use LWP::UserAgent;
 use URI::Escape;
+use HTTP::Request;
+use Carp;
 
 # debugging
 use Data::Dumper;
 
-our $VERSION = '0.01';
+our $VERSION = '0.02';
 our $errstr = '';
 
 =head1 NAME
@@ -23,25 +25,35 @@ WWW::Metaweb - An interface to the Metaweb database via MQL
   use strict;
   use WWW::Metaweb;
 
-  my $mh = WWW::Metaweb->new( username => $u,
-  			      password => $p, 
-			      server => 'www.freebase.com',
-			      auth_uri => '/api/account/login',
-			      read_uri => '/api/service/mqlread',
-			      write_uri => '/api/service/mqlwrite',
-			      trans_uri => '/api/trans',
-			      pretty_json => 1 );
+  my $mh = WWW::Metaweb->connect( username => $u,
+				  password => $p, 
+				  server => 'www.freebase.com',
+				  auth_uri => '/api/account/login',
+				  read_uri => '/api/service/mqlread',
+				  write_uri => '/api/service/mqlwrite',
+				  trans_uri => '/api/trans',
+				  pretty_json => 1 );
 
   my $query = {
-	  name => 'Nico Minoru',
+	  '/type/object/creator' => undef,
+	  cover_appearances => [{
+	    type => '/comic_books/comic_book_issue',
+	    name => undef,
+	    part_of_series => undef
+	  }],
+	  created_by => [],
 	  id => undef,
-	  type => [],
-	  '/comic_books/comic_book_character/cover_appearances' => [{
-		  name => undef,
-		  id => undef,
-	  }]
+	  name => 'Nico Minoru',
+	  type => '/comic_books/comic_book_character'
   };
 
+The easy way:
+
+  my $result = $mh->read($query, 'json');
+  print $result;
+
+The complicated way:
+
   $mh->add_query('read', $query);
   $mh->send_envelope('read')
     or die $WWW::Metaweb::errstr;
@@ -51,7 +63,7 @@ WWW::Metaweb - An interface to the Metaweb database via MQL
 
 =head1 ABSTRACT
 
-Metaweb provides an interface to a Metaweb database through it's HTTP API and MQL.
+WWW::Metaweb provides an interface to a Metaweb database through it's HTTP API and MQL.
 
 =head1 DESCRIPTION
 
@@ -63,7 +75,13 @@ Qeueries are written in the Metaweb Query Language (MQL), using Javascript Objec
 
 =head2 Class methods
 
-=head3 $version = WWW::Metaweb->version
+=over
+
+=item B<< $version = WWW::Metaweb->version >>
+
+Returns the version of WWW::Metaweb being used.
+
+=back
 
 =cut
 
@@ -73,11 +91,53 @@ sub version  {
 
 =head2 Constructors
 
-=cut
+=over
 
-=head3 $mh = WWW::Metaweb->connect(username => $u, password => $p [, option_key => 'option_value'])
+=item B<< $mh = WWW::Metaweb->connect( [option_key => 'option_value' ...] ) >>
 
-While this method no longer requires a username and password, if they are supplied by are incorrect (or Metaweb can't authenticate for any other reason) then C<undef> will be returned.
+Returns a new WWW::Metaweb instance, a number of different attributes can be sethere (see below).
+
+If a C<username> and C<password> are supplied then C<connect()> will attempt to authenticate before returning. If this authentication fails then C<undef> will be returned.
+
+=over
+
+=item B<< Metaweb parameters >>
+
+=over
+
+=item B<< auth_uri >>
+
+The URI used to authenticate for this Metaweb (eg. /api/account/login).
+
+=item B<< read_uri >>
+
+The URI used to submit a read MQL query to this Metaweb (eg. /api/service/mqlread).
+
+=item B<< write_uri >>
+
+The URI used to submit a write MQL query to this Metaweb (eg. /api/service/mqlwrite).
+
+=item B<< trans_uri >>
+
+The URI used to access the translation service for this Metaweb (eg. /api/trans). Please note this this URI does not include the actual C<translation>, at this time these are C<raw>, C<image_thumb> and C<blurb>.
+
+=back
+
+=item B<< JSON parameters >>
+
+=over
+
+=item B<< pretty_json >>
+
+Determines whether the response to a JSON query is formatted nicely. This is just passed along to the JSON object as C<JSON::XS->new->pretty($mh->{pretty})>.
+
+=item B<< json_preprocessor >>
+
+Can provide a reference to a sub-routine that pre-processes JSON queries, the sub-routine should expect one argument - the JSON query as a string and return the processed JSON query as a scalar.
+
+=back
+
+=back
 
 =cut
 
@@ -88,8 +148,10 @@ sub connect  {
 
 
 	my $options = { @_ };
-	$username = $options->{username};
-	$password = $options->{password};
+	$username = $options->{username} || '' if exists $options->{username};
+	$password = $options->{password} || '' if exists $options->{password};
+	delete $options->{username};
+	delete $options->{password};
 
 	my $self = {
 		     auth_uri => undef,
@@ -99,16 +161,25 @@ sub connect  {
 		     read_envelope => { },
 		     write_envelope => { },
 		     result_envelope => { },
-		     pretty_json => 0
+		     json_preprocessor => undef,
+		     pretty_json => 0,
+		     query_counter => 0,
+		     debug => 0
 		   };
 	
 	bless $self, $class;
 	
+	$self->server($options->{server}); # Sets the server.
+	delete $options->{server};
 	# Sets the option attributes from $options into $self.
 	foreach my $key (keys %$options)  {
-		$self->{$key} = $options->{$key} if (exists $self->{$key});
+		if (exists $self->{$key})  {
+			$self->{$key} = $options->{$key};
+		}
+		else  {
+			carp "Unknown option '$key' used in connect().";
+		}
 	}
-	$self->server($options->{server}); # Sets the server.
 
 	# A little bit of vanity here (the agent).
 	$self->useragent(LWP::UserAgent->new( agent => 'Metaweb/'.$WWW::Metaweb::VERSION,
@@ -124,9 +195,13 @@ sub connect  {
 	return $self;
 } # ->connect
 
+=back
+
 =head2 Authentication
 
-=head3 $mh->authenticate($username, $password)
+=over
+
+=item B<< $mh->authenticate($username, $password) >>
 
 Authenticates to the auth_uri using the supplied username and password. If the authentication is successful then the cookie is retained for future queries.
 
@@ -153,8 +228,8 @@ sub authenticate  {
 
 	unless ($raw_header = $response->header('Set_Cookie'))  {
 		# Authentication failed.
-		my $json = JSON->new;
-		my $reply = $json->jsonToObj($response->content);
+		my $jsonxs = JSON::XS->new->utf8;
+		my $reply = $jsonxs->decode($response->content);
 		$WWW::Metaweb::errstr = "Login failed: [status: $reply->{status}, code: $reply->{code}]";
 		
 		return undef;
@@ -174,9 +249,184 @@ sub authenticate  {
 	return 1;
 } # ->authenticate
 
-=head2 Query manipulation
+=back
 
-=head3 $mh->add_query($method, query_name1 => $query1 [, query_name2 => $query2 [, ...]])
+=head2 Easy Querying
+
+=over
+
+=item B<< @results = $mh->read($read_query [, $read_query2 ...] [, $format]) >> or B<< $result = $mh->read($read_query [, $format]) >>
+
+The easy way to perform a read query.
+
+Accepts one or more queries which are bundled up in one envelope and sent to the read service. The response is an array containing the results in the same order as the queries were given in.
+
+If only one query is given and assigned to a scaler then the single query will be returned as a scaler instead of in an array.
+
+=cut
+
+sub read  {
+	my $self = shift;
+	my @read_queries = @_;
+	my ($i, $format);
+
+	$self->clear_read_queries;
+	
+	# Add each query to the envelope and replace it's place in the array
+	# with the query name asigned to it.
+	for ($i = 0; $i < @read_queries; $i++)  {
+		if ($read_queries[$i] eq 'perl' || $read_queries[$i] eq 'json')  {
+			$format = $read_queries[$i];
+			delete $read_queries[$i];
+		}
+		else  {
+			my $read_query = $read_queries[$i];
+			$read_queries[$i] = "query$i";
+
+			$self->add_read_query($read_queries[$i] => $read_query);
+			#carp 'WWW::Metaweb - Bad format in read() - ($format = \'' . $read_queries[$i] . '\')';
+			#delete $read_queries[$i];
+		}
+	}
+
+	# We're helpless if this fails, return undef and trust the errstr has
+	# been set further down.
+	return undef unless (defined $self->send_read_envelope);
+
+	# Replace the query names in our array with the result of the query.
+	map { $_ = $self->result($_, $format); } @read_queries;
+
+	# If there is only one result and an array hasn't been asked for, return
+	# the single value as a scaler instead.
+	return (@read_queries == 1 && not wantarray) ? $read_queries[0] : @read_queries;
+} # ->read
+
+=item B<< @result = $mh->write($write_query [, $write_query2 ...] [, $format]) >> or B<< $result = $mh->write($write_query [, $format]) >>
+
+The easy way to perform a write query.
+
+The syntax and behaviour are exactly the same as C<read()> (above).
+
+=cut
+
+sub write  {
+	my $self = shift;
+	my @write_queries = @_;
+	my ($i, $format);
+
+	# This method works exactly the same as the read method.
+	
+	$self->clear_write_queries;
+	
+	for ($i = 0; $i < @write_queries; $i++)  {
+		if ($write_queries[$i] eq 'perl' || $write_queries[$i] eq 'json')  {
+			$format = $write_queries[$i];
+			delete $write_queries[$i];
+		}
+		else  {
+			my $write_query = $write_queries[$i];
+			$write_queries[$i] = "query$i";
+
+			$self->add_write_query($write_queries[$i] => $write_query);
+		}
+
+	}
+
+	return undef unless defined $self->send_write_envelope;
+
+	map { $_ = $self->result($_, $format); } @write_queries;
+
+	return (@write_queries == 1 && not wantarray) ? $write_queries[0] : @write_queries;
+} # ->write
+
+=back
+
+=head2 Translation Service
+
+=over
+
+=item B<< $content = $mh->trans($translation, $guid) >>
+
+Gets the content for a C<guid> in the format specified by C<$translation>. Metaweb currently supports the translations C<raw>, C<image_thumb> and C<blurb>.
+
+C<$translation> is not checked for validity, but an error will most likely be returned by the server.
+
+C<$guid> should be the global identifier of a Metaweb object of type C</common/image> or C</type/content> and/or C</common/document> depending on the translation requested, if not the Metaweb will return an error. The global identifier can be prefixed with either a '#' or the URI escaped version '%23' then followed by the usual string of lower case hex.
+
+=cut
+
+sub trans  {
+	my $self = shift;
+	my $translation = shift;
+	my $guid = lc shift;
+	my ($url, $response);
+
+	# Check that the guid looks mostly correct and replace a hash at the
+	# beginning of the guid with the URI escape code.
+	unless ($guid =~ s/^(\#|\%23)([\da-f]+)$/\%23$2/)  {
+		$WWW::Metaweb::errstr = "Bad guid: $guid";
+		return undef;
+	}
+
+	$url = $self->server.$self->{trans_uri}.'/'.$translation.'/'.$guid;
+	$response = $self->useragent->get($url);
+	
+	# An HTTP response that isn't success indicates something bad has
+	# happened and there's nothing I can do about it.
+	unless ($response->is_success)  {
+		$WWW::Metaweb::errstr = "Trans query failed, HTTP response: " . $response->status_line;
+		return undef;
+	}
+
+	return $response->content;
+} # ->trans
+
+=item B<< $content = $mh->raw($guid) >>
+
+Convenience method for getting a C<raw> translation of the object with C<$guid>. See C<trans()> for more details.
+
+=cut
+
+sub raw  {
+	my $self = shift;
+	my $guid = shift;
+
+	return $self->trans('raw', $guid);
+} # ->raw
+
+=item B<< $content = $mh->image_thumb($guid) >>
+
+Convenience method for getting a C<image_thumb> translation of the object with C<$guid>. See C<trans()> for more details.
+
+=cut
+
+sub image_thumb  {
+	my $self = shift;
+	my $guid = shift;
+
+	return $self->trans('image_thumb', $guid);
+} # ->image_thumb
+
+=item B<< $content = $mh->blurb($guid) >>
+
+Convenience method for getting a C<blurb> translation of the object with C<$guid>. See C<trans()> for more details.
+
+=cut
+
+sub blurb  {
+	my $self = shift;
+	my $guid = shift;
+
+	return $self->trans('blurb', $guid);
+} # ->blurb
+
+=back
+
+=head2 Complicated Querying
+
+=over
+
+=item B<< $mh->add_query($method, query_name1 => $query1 [, query_name2 => $query2 [, ...]]) >>
 
 This method adds queries to a query envelope. C<$method> must have a value of either 'read' or 'write'.
 
@@ -240,16 +490,40 @@ sub add_query  {
 		return undef;
 	}
 
+	my ($query_name, $query);
 	my $no_error = 1;	
-	foreach my $query_name (keys %$queries)  {
+	foreach $query_name (keys %$queries)  {
+		$query = $queries->{$query_name};
 		$no_error = 0 unless $self->check_query_syntax($method, $queries->{$query_name});
-		$self->{$envelope}->{$query_name} = $self->perl_query($queries->{$query_name});
+
+		if (ref $query eq 'HASH' or ref $query eq 'ARRAY')  {
+			# It's a Perl structure.
+			if (((ref $query) eq 'HASH' && (not defined $query->{query})) || ((ref $query) eq 'ARRAY' && (not defined $query->[0]->{query})))  {
+				$query = { query => $query };
+			}
+			$query->{anti_cache} = (time . $self->{query_counter}++);
+			$query->{cursor} = JSON::XS::true if $self->{auto_cursors} && not defined $query->{cursor};
+		}
+		elsif ((not ref $query))  {
+			# It's a JSON string - but we'll convert it to Perl and
+			# back again to manipulate it.
+			my ($jxs, $p_query);
+			$p_query = from_json($query);
+			if (((ref $p_query) eq 'HASH' && (not defined $p_query->{query})) || ((ref $p_query eq 'ARRAY') && (not defined $p_query->[0]->{query})))  {	
+				$p_query = { query => $p_query };
+			}
+			$p_query->{anti_cache} = (time . $self->{query_counter}++);
+			$p_query->{cursor} = JSON::XS::true if $self->{auto_cursors} && not defined $query->{cursor};
+			$query = to_json($p_query);
+		}
+		# Now store it for sending.
+		$self->{$envelope}->{$query_name} = $query;
 	}
 
 	return $no_error;
 } # ->add_query
 
-=head3 $mh->clear_queries($method)
+=item B<< $mh->clear_queries($method) >>
 
 Clears all the previous queries from the envelope.
 
@@ -269,7 +543,7 @@ sub clear_queries  {
 	return 1;
 } # ->clear_queries
 
-=head3 $count = $mh->query_count($method)
+=item B<< $count = $mh->query_count($method) >>
 
 Returns the number of queries held in the C<$method> query envelope.
 
@@ -282,13 +556,12 @@ sub query_count  {
 
 	return undef unless $envelope = __test_envelope($method, 'query_count');
 	@keys = keys %{$self->{$envelope}};
-	#print Dumper $self->{$envelope};
 	$key_count = @keys;
 	
 	return $key_count; 
 } # ->query_count
 
-=head3 $bool = $mh->check_query_syntax($method, $query)
+=item B<< $bool = $mh->check_query_syntax($method, $query) >>
 
 Returns a boolean value to indicate whether the query provided (either as a Perl structure or a JSON string) follows correct MQL syntax. C<$method> should be either 'read' or 'write' to indicate which syntax to check query against.
 
@@ -304,41 +577,7 @@ sub check_query_syntax  {
 	return 1;
 } # ->check_query_syntax
 
-=head3 $obj = $mh->perl_query($query)
-
-Forces a query into a Perl structure if it's a JSON string. If given a Perl structure it is left more or less alone.
-
-=cut
-
-sub perl_query  {
-	my $self = shift;
-	my $original_query = shift;
-	my ($format, $json_string, $perl_query);
-	my $json = JSON->new;
-
-	if (ref $original_query)  {
-		$format = 'perl';
-		$json_string = $json->objToJson($original_query);
-	}
-	else  {
-		$format = 'json';
-		$json_string = $original_query;
-
-	}
-
-	$perl_query = {
-			query => $json->jsonToObj($json_string),
-			format => $format
-		     };
-	
-	if (keys %{$perl_query->{query}} == 1 && defined $perl_query->{query}->{query})  {
-		$perl_query->{query} = $perl_query->{query}->{query};
-	}
-
-	return $perl_query;
-} # ->perl_query
-
-=head3 $http_was_successful = $mh->send_envelope($method)
+=item B<< $http_was_successful = $mh->send_envelope($method) >>
 
 Sends the current query envelope and returns whether the HTTP portion was successful. This does not indicate that the query itself was well formed or correct.
 
@@ -353,14 +592,57 @@ sub send_envelope  {
 
 	return undef unless $envelope = __test_envelope($method, 'send_envelope');
 
-	my $json = JSON->new;
-	my ($url, $response);
-	my $json_envelope = $json->objToJson($self->{$envelope});
-	$json_envelope =~ s/"format":"(?:json|perl),"//g;
+	my $jsonxs = JSON::XS->new->utf8;
+	my ($json_envelope, $url, $request, $response);
 
-	$url = $self->server.$self->{$method.'_uri'}.'?queries='.uri_escape($json_envelope);
+	# Create a list of pre-processors
+	my @preprocessors;
+	if (ref $self->{json_preprocessor} eq 'CODE')  {
+		@preprocessors = ( $self->{json_preprocessor} );
+	}
+	elsif (ref $self->{json_preprocessor} eq 'ARRAY')  {
+		foreach my $sub (@{$self->{json_preprocessor}})  {
+			push @preprocessors, $sub if (ref $sub eq 'CODE');
+		}
+	}
+	
+	my $first = 1;
+	$json_envelope = '{';
+	foreach my $query_name (keys %{$self->{$envelope}})  {
+		my $query = (ref $self->{$envelope}->{$query_name}) ? $jsonxs->encode($self->{$envelope}->{$query_name}) : $self->{$envelope}->{$query_name};
+		#$query =~ s/"format":"(?:json|perl)",//;
 
-	$response = $self->useragent->get($url);
+		foreach my $sub (@preprocessors)  {
+			$query = &$sub($query);
+		}
+		# If the query has been botched - set it to an empty string.
+		$query = '' unless defined $query;
+
+		$json_envelope .= ',' if $first == 0;
+		$json_envelope .= '"'.$query_name.'":'.$query;
+
+		$first = 0;
+	}
+	$json_envelope .= '}';
+
+	print $json_envelope . "\n" if $self->{debug};
+	
+	# Set up the request depending on whether this is a read or write op.
+	$request = HTTP::Request->new;
+	$request->header( 'X-Metaweb-Request' => 'True' );
+	if ($method eq 'read')  {
+		$request->method('GET');
+		$request->uri($self->server.$self->{$method.'_uri'}.'?queries='.uri_escape($json_envelope));
+	}
+	else  {
+		$request->method('POST');
+		$request->uri($self->server.$self->{$method.'_uri'});
+		$request->content_type('application/x-www-form-urlencoded'); 
+		$request->content('queries='.uri_escape($json_envelope));
+	}
+	$response = $self->useragent->request($request);
+
+	$self->{last_envelope_sent} = $method;
 
 	unless ($response->is_success)  {
 		$WWW::Metaweb::errstr = "Query failed, HTTP response: " . $response->status_line;
@@ -370,15 +652,21 @@ sub send_envelope  {
 	return ($self->set_result($method, $response->content)) ? $response->is_success : undef;
 } # ->send_envelope
 
-=head2 Query Convenience Methods
+=back
+
+=head2 Query Convenience Methods (for complicated queries)
 
 As most of the query and result methods require a C<$method> argument as the first parameter, I've included methods to call them for each method explicitly.
 
 If you know that you will always be using a method call for either a read or a write query/result, then it's safer to user these methods as you'll get a compile time error if you spell read or write incorrectly (eg. a typo), rather than a run time error.
 
-=head3 $mh->add_read_query(query_name1 => $query1 [, query_name2 => $query2 [, ...]])
+Of course it's probably much easier to just use C<read()> and C<write()> from the L<Easy Querying> section above.
 
-Convenience method to add a read query. See add_query() for details.
+=over
+
+=item B<< $mh->add_read_query(query_name1 => $query1 [, query_name2 => $query2 [, ...]]) >>
+
+Convenience method to add a read query. See C<add_query()> for details.
 
 =cut
 
@@ -388,9 +676,9 @@ sub add_read_query  {
 	return $self->add_query('read', @_);
 } # ->add_read_query
 
-=head3 $mh->add_write_query(query_name1 => $query1 [, query_name2 => $query2 [, ...]])
+=item B<< $mh->add_write_query(query_name1 => $query1 [, query_name2 => $query2 [, ...]]) >>
 
-Convenience method to add a write query. See add_query() for details.
+Convenience method to add a write query. See C<add_query()> for details.
 
 =cut
 
@@ -400,9 +688,9 @@ sub add_write_query  {
 	return $self->add_query('write', @_);
 } # ->add_write_query
 
-=head3 $mh->clear_read_queries
+=item B<< $mh->clear_read_queries >>
 
-Convenience method to clear the read envelope. See clear_queries() for details.
+Convenience method to clear the read envelope. See C<clear_queries()> for details.
 
 =cut
 
@@ -412,9 +700,9 @@ sub clear_read_queries  {
 	return $self->clear_queries('read', @_);
 } # ->clear_read_queries
 
-=head3 $mh->clear_write_queries
+=item B<< $mh->clear_write_queries >>
 
-Convenience method to clear the write envelope. See clear_queries() for details.
+Convenience method to clear the write envelope. See C<clear_queries()> for details.
 
 =cut
 
@@ -424,9 +712,9 @@ sub clear_write_queries  {
 	return $self->clear_queries('write', @_);
 } # ->clear_write_queries
 
-=head3 $count = $mh->read_query_count
+=item B<< $count = $mh->read_query_count >>
 
-Convenience method, returns the number of queries in the read envelope. See query_count() for details.
+Convenience method, returns the number of queries in the read envelope. See C<query_count()> for details.
 
 =cut
 
@@ -436,9 +724,9 @@ sub read_query_count  {
 	return $self->query_count('read', @_);
 } # ->read_query_count
 
-=head3 $count = $mh->write_query_count
+=item B<< $count = $mh->write_query_count >>
 
-Convenience method, returns the number of queries in the write envelope. See query_count() for details.
+Convenience method, returns the number of queries in the write envelope. See C<query_count()> for details.
 
 =cut
 
@@ -448,9 +736,9 @@ sub write_query_count  {
 	return $self->query_count('write', @_);
 } # ->write_query_count
 
-=head3 $http_was_successful = $mh->send_read_envelope
+=item B<< $http_was_successful = $mh->send_read_envelope >>
 
-Convenience method, sends the read envelope. See send_envelope() for details.
+Convenience method, sends the read envelope. See C<send_envelope()> for details.
 
 =cut
 
@@ -460,9 +748,9 @@ sub send_read_envelope  {
 	return $self->send_envelope('read');
 } # ->send_read_envelope
 
-=head3 $http_was_successful = $mh->send_write_envelope
+=item B<< $http_was_successful = $mh->send_write_envelope >>
 
-Convenience method, sends the write envelope. See send_envelope() for details.
+Convenience method, sends the write envelope. See C<send_envelope()> for details.
 
 =cut
 
@@ -472,10 +760,13 @@ sub send_write_envelope  {
 	return $self->send_envelope('write');
 } # ->send_write_envelope
 
+=back
 
-=head2 Result manipulation
+=head2 Result manipulation (for complicated queries)
 
-=head3 $mh->set_result($json)
+=over
+
+=item B<< $mh->set_result($json) >>
 
 Sets the result envelope up so that results can be accessed for the latest query. Any previous results are destroyed.
 
@@ -487,27 +778,28 @@ sub set_result  {
 	my $self = shift;
 	my $method = shift;
 	my $json_result = shift;
-	my $json = JSON->new;
 	my $envelope;
 
 	return undef unless $envelope = __test_envelope($method, 'set_result');
 	
-	$self->{result_envelope} = $json->jsonToObj($json_result);
+	$self->{result_envelope} = $json_result;
+	my $perl_result = from_json($json_result);
 
-	# Potential bug here if an error occurs
-	foreach my $query_name (keys %{$self->{result_envelope}})  {
-		$self->{result_envelope}->{$query_name}->{format} = $self->{$envelope}->{$query_name}->{format} unless ($query_name eq 'status' || $query_name eq 'code');
+	my $status = $perl_result->{status};
+	unless ($status eq '200 OK')  {
+		$WWW::Metaweb::errstr = 'Bad outer envelope status: ' . $status;
+		return 0;
 	}
 
-	unless ($self->{result_envelope}->{status} eq '200 OK')  {
-		$WWW::Metaweb::errstr = 'Bad outer envelope status: ' . $self->{result_envelope}->{status};
-		return 0;
+	$self->{result_format} = { };
+	foreach my $query_name (keys %{$self->{$envelope}})  {
+		$self->{result_format}->{$query_name} = (ref $self->{$envelope}->{$query_name}) ? 'perl' : 'json';
 	}
 
 	return 1;
 } # ->set_result
 
-=head3 $bool = $mh->result_is_ok($query_name)
+=item B<< $bool = $mh->result_is_ok($query_name) >>
 
 Returns a boolean result indicating whether the query named C<$query_name> returned a status ok. Returns C<undef> if there is no result for C<query_name>.
 
@@ -518,24 +810,28 @@ sub result_is_ok  {
 	my $query_name = shift || 'netmetawebquery';
 	my $result_is_ok = undef;
 
-	if (defined $self->{result_envelope}->{$query_name})  {
-		if ($self->{result_envelope}->{$query_name}->{code} eq '/api/status/ok')  {
+	my $result = from_json($self->{result_envelope})->{$query_name};	
+	if (defined $result)  {
+		my ($code, $message);
+		$code = $result->{code};
+		if ($code eq '/api/status/ok')  {
 			$result_is_ok = 1;
 		}
 		else  {
-			$WWW::Metaweb::errstr = 'Result status not okay: ' . $self->{result_envelope}->{$query_name}->{code};
-			$result_is_ok =  0;
+			$message = $result->{messages}->[0]->{message};
+			$WWW::Metaweb::errstr = "Result status not okay for $query_name: $code; error: $message;";
 		}
+
 	}
 	else  {
 		$WWW::Metaweb::errstr = 'No result found for query name: ' . $query_name;
 		$result_is_ok = undef;
 	}
-
+	
 	return $result_is_ok;
 } # ->result_is_okay
 
-=head3 $mh->result($query_name [, $format])
+=item B<< $mh->result($query_name [, $format]) >>
 
 Returns the result of query named C<$query_name> in the format C<$format>, which should be either 'perl' for a Perl structure or 'json' for a JSON string.
 
@@ -560,109 +856,53 @@ sub result  {
 	my $query_name = shift || 'netmetawebquery';
 	my $format = shift;
 	my $result;
+	my $raw_result;
+	my $perl_result;
 
-	unless (defined $self->{result_envelope}->{$query_name})  {
-		$WWW::Metaweb::errstr = "No result found with the name: $query_name";
-		return undef;
-	}
-	$format = $self->{result_envelope}->{$query_name}->{format} unless defined $format;
+	# If the query isn't okay - just return undef, errstr will have been set
+	return undef unless $self->result_is_ok($query_name);
+
+	# Check the return format if it hasn't been explicitly set.
+	$format = $self->{result_format}->{$query_name} unless defined $format;
+	
+	$JSON::UnMapping = 1;
+	$perl_result = from_json($self->{result_envelope})->{$query_name}->{result};
 
 	if ($format eq 'json')  {
-		my $json = JSON->new;
-		$result = $json->objToJson($self->{result_envelope}->{$query_name}->{result}, { pretty => $self->{pretty_json} });
+		$result = JSON::XS->new->utf8->pretty($self->{pretty_json})->encode($perl_result);
 	}
 	else  {
-		$result = $self->{result_envelope}->{$query_name}->{result};
+		$result = $perl_result;
 	}
-
+	
 	return $result;
 } # ->result
 
-=head2 Translations
+=item B<< $text = $mh->raw_result >>
+
+Returns the raw result from the last time an envelope was sent.
+
+After a successful query this will most likely be a JSON structure consisting of the outer envelope with the code and status as well as a result for each query sent in the last batch.
+
+After an unsuccessful query this will contain error messages detailing what went wrong as well as code and status sections to similar effect.
+
+If the transaction itself failed then the returned text will probably be empty, but at the very least this method will always return an empty string, never C<undef>.
 
 =cut
 
-=head3 $content = $mh->trans($translation, $guid)
-
-Gets the content for a C<guid> in the format specified by C<$translation>. WWW::Metaweb currently supports the translations C<raw>, C<image_thumb> and C<blurb>.
-
-C<$translation> is not checked for validity, but an error will most likely be returned by the server.
-
-C<$guid> should be the global identifier of a Metaweb object of type C</common/image> or C</type/content> and/or C</common/document> depending on the translation requested, if not the Metaweb will return an error. The global identifier can be prefixed with either a '#' or the URI escaped version '%23' then followed by the usual string of lower case hex.
-
-=cut
-
-sub trans  {
+sub raw_result  {
 	my $self = shift;
-	my $translation = shift;
-	my $guid = lc shift;
-	my ($url, $response);
 
-	# Check that the guid looks mostly correct and replace a hash at the
-	# beginning of the guid with the URI escape code.
-	unless ($guid =~ s/^(\#|\%23)([\da-f]+)$/\%23$2/)  {
-		$WWW::Metaweb::errstr = "Bad guid: $guid";
-		return undef;
-	}
+	return $self->{result_envelope} || '';
+} # ->raw_result
 
-	$url = $self->server.$self->{trans_uri}.'/'.$translation.'/'.$guid;
-	$response = $self->useragent->get($url);
-	
-	# An HTTP response that isn't success indicates something bad has
-	# happened and there's nothing I can do about it.
-	unless ($response->is_success)  {
-		$WWW::Metaweb::errstr = "Trans query failed, HTTP response: " . $response->status_line;
-		return undef;
-	}
-
-	return $response->content;
-} # ->trans
-
-=head3 $content = $mh->raw($guid)
-
-Convenience method for getting a C<raw> translation of the object with C<$guid>. See C<trans> for more details.
-
-=cut
-
-sub raw  {
-	my $self = shift;
-	my $guid = shift;
-
-	return $self->trans('raw', $guid);
-} # ->raw
-
-=head3 $content = $mh->image_thumb($guid)
-
-Convenience method for getting a C<image_thumb> translation of the object with C<$guid>. See C<trans> for more details.
-
-=cut
-
-sub image_thumb  {
-	my $self = shift;
-	my $guid = shift;
-
-	return $self->trans('image_thumb', $guid);
-} # ->image_thumb
-
-=head3 $content = $mh->blurb($guid)
-
-Convenience method for getting a C<blurb> translation of the object with C<$guid>. See C<trans> for more details.
-
-=cut
-
-sub blurb  {
-	my $self = shift;
-	my $guid = shift;
-
-	return $self->trans('blurb', $guid);
-} # ->blurb
+=back
 
 =head2 Accessors
 
-=cut
+=over
 
-=head3 $ua = $mh->useragent
-=head3 $mh->useragent($ua)
+=item B<< $ua = $mh->useragent >> or B<< $mh->useragent($ua) >>
 
 Gets or sets the LWP::UserAgent object which is used to communicate with the Metaweb. This method can be used to change the user agent settings (eg. C<$mh->useragent->timeout($seconds)>).
 
@@ -677,8 +917,7 @@ sub useragent  {
 	return $self->{ua};
 } # ->useragent
 
-=head3 $host = $mh->server
-=head3 $mh->server($new_host)
+=item B<< $host = $mh->server >> or B<< $mh->server($new_host) >>
 
 Gets or sets the host for this Metaweb (eg. www.freebase.com). No checking is currently done as to the validity of this host.
 
@@ -694,34 +933,7 @@ sub server  {
 	return $self->{server};
 } # ->server
 
-
-=head1 ATTRIBUTES
-
-The following attributes can be set manually by the programmer.
-
-=head2 WWW::Metaweb set-up
-
-=head3 auth_uri
-
-The URI used to authenticate for this Metaweb (eg. /api/account/login).
-
-=head3 read_uri
-
-The URI used to submit a read MQL query to this Metaweb (eg. /api/service/mqlread).
-
-=head3 write_uri
-
-The URI used to submit a write MQL query to this Metaweb (eg. /api/service/mqlwrite).
-
-=head3 trans_uri
-
-The URI used to access the translation service for this Metaweb (eg. /api/trans). Please note this this URI does not include the actual C<translation>, at this time these are C<raw>, C<image_thumb> and C<blurb>.
-
-=head2 Formatting
-
-=head3 pretty_json
-
-Determines whether the response to a JSON query is formatted nicely. This is just passed along to the JSON object as C<JSON->new(pretty => $mh->{pretty})>.
+=back
 
 =head1 BUGS AND TODO
 
@@ -739,6 +951,8 @@ I would like to implement transparent cursors in read queries so a single query
 
 While entirely rewritten, I think it's only fair to mention that the basis for the core of this code is the Perl example on Freebase (http://www.freebase.com/view/helptopic?id=%239202a8c04000641f800000000544e139).
 
+Michael Jones has also been a great help - pointing out implementation issues and providing suggested fixes and code.
+
 =head1 SEE ALSO
 
 Freebase, Metaweb
@@ -776,8 +990,6 @@ sub __test_envelope  {
 	return $envelope;
 } # &__test_envelope
 
-
-
 return 1;
 __END__
 
